/*
 * debug - main.c
 * 	version 0.0.4
 * 	
 * Copyright (C) 2005..2010  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_LIBREADLINE
#include <readline/readline.h>
#include <readline/history.h>
#endif
#include <errno.h>
#include <debug_io.h>
#include <debug_mem.h>
#include <debug_msr.h>
#include <getopt.h>
#include <string.h>
#include <unistd.h>
#ifdef HAVE_LIBPCI
#include <pci/pci.h>
#endif

#define PROMPT	"> "

static int ac_compat = 0;	/* Linux MiniDEBUG compatibility mode */
static int parambase = 16;	/* used in Linux MiniDEBUG compatibility mode */
static int dump_align = 0;
static int rel_mem = 0;
static int rel_io = 0;
static unsigned long membase = 0;
static unsigned long iobase = 0;
static int hide_nonascii = 1;
static int hide_nonprintable = 1;
static char *run_script = NULL;
static int io_access_mode = IOA_DEFAULT;
static int quiet_mode;
static int pci_domain, pci_bus, pci_dev, pci_fn;

#define USE_READLINE_NO		0
#define USE_READLINE_DEFAULT	1
#define USE_READLINE_YES	2

static int use_readline = USE_READLINE_DEFAULT;

static int parse_error(const char *fmt)
{
	if (!quiet_mode)
		fprintf(stderr, "parse error: %s\n", fmt);
	return 1;
}

char *std_gets(void)
{
	char *tmp;

	if (feof(stdin))
		return NULL;
	tmp = malloc(1024);
	if (tmp == NULL) {
		fprintf(stderr, "fatal: Can't allocate memory.\n");
		exit(1);
	}
	if (fgets(tmp, 1024, stdin) == NULL) {
		free(tmp);
		return NULL;
	}
	return tmp;
}

char *rl_gets(void)
{
#ifdef HAVE_LIBREADLINE
	char *tmp;

	tmp = readline(PROMPT);
	if (tmp && *tmp)
		add_history(tmp);
	return tmp;
#else
	fprintf(stderr, "> ");
	return std_gets();
#endif
}

void print_int8(uint8_t value)
{
	printf("0x%02hhx (%hhi)\n", value, value);
}
	
void print_int16(uint16_t value)
{
	printf("0x%04hx (%hi)\n", value, value);
}

void print_int32(uint32_t value)
{
	printf("0x%08x (%i)\n", value, value);
}

void print_int64(uint64_t value)
{
	printf("0x%016" PRIx64 " (%" PRId64 ")\n", value, value);
}

unsigned long gen_stoul(const char *str)
{
	if (ac_compat) {
		if (str[0] == '$')
			return (unsigned long) strtoll(str + 1, NULL, 16);
		if (str[0] == '#')
			return (unsigned long) strtoll(str + 1, NULL, 10);
		return (unsigned long) strtoll(str, NULL, parambase);
	}
	return (unsigned long) strtoll(str, NULL, 0);
}

unsigned long long gen_stoull(const char *str)
{
	if (ac_compat) {
		if (str[0] == '$')
			return strtoull(str + 1, NULL, 16);
		if (str[0] == '#')
			return strtoull(str + 1, NULL, 10);
		return strtoull(str, NULL, parambase);
	}
	return strtoull(str, NULL, 0);
}

int gen_parse_1ul(const char *args, unsigned long *addr)
{
	char cmd[64], addr_s[16];
	int ret;
	
	ret = sscanf(args, "%63s %15s", cmd, addr_s);
	if (ret < 2)
		return 1;
	*addr = gen_stoul(addr_s);
	return 0;
}

int gen_parse_2ul(const char *args, unsigned long *addr, unsigned long *value)
{
	char cmd[64], addr_s[16], value_s[16];
	int ret;
	
	ret = sscanf(args, "%63s %15s %15s", cmd, addr_s, value_s);
	if (ret < 3)
		return 1;
	*addr = gen_stoul(addr_s);
	*value = gen_stoul(value_s);
	return 0;
}

int gen_parse_ul_ull(const char *args, unsigned long *addr,
		unsigned long long *value)
{
	char cmd[64], addr_s[15], value_s[32];
	int ret;
	
	ret = sscanf(args, "%63s %15s %31s", cmd, addr_s, value_s);
	if (ret < 3)
		return 1;
	*addr = gen_stoul(addr_s);
	*value = gen_stoull(value_s);
	return 0;
}

int gen_parse_3ul(const char *args, unsigned long *addr, unsigned long *value,
		unsigned long *value2)
{
	char cmd[64], addr_s[24], value_s[24], value2_s[24];
	int ret;
	
	ret = sscanf(args, "%63s %23s %23s %23s", cmd, addr_s, value_s,
			value2_s);
	if (ret < 4)
		return 1;
	*addr = gen_stoul(addr_s);
	*value = gen_stoul(value_s);
	*value2 = gen_stoul(value2_s);
	return 0;
}

static char c_to_pc(uint8_t c)
{
	if ((hide_nonprintable && (c < 32)) || (hide_nonascii && (c > 127)))
		return '.';
	return c;
}

void dump_area_a(unsigned long base, uint8_t *buff, ssize_t len)	
{
	unsigned long i, j;

	len += base & 0xf;
	for (i = base & ~0xf; len > 0; i += 16, len -= 16) {
		printf("0x%08lx: ", i);
		for (j = 0; j < 16; j++) {
			if (j == 8)
				printf(" ");

			if ((i + j < base) || (j >= len))
				printf("   ");
			else
				printf("%02hhx ", buff[i + j - base]);
		}
		printf(" ");
		for (j = 0; j < 16; j++) {
			if ((i + j < base) || (j >= len))
				printf(" ");
			else
				printf("%c", c_to_pc(buff[i + j - base]));
		}
		printf("\n");
	}
}

void dump_area(unsigned long base, uint8_t *buff, ssize_t len)	
{
	unsigned long i, j;

	for (i = base; len > 0; i += 16, len -= 16) {
		printf("0x%08lx: ", i);
		for (j = 0; j < 16; j++) {
			if (j == 8)
				printf(" ");

			if (j >= len)
				printf("   ");
			else
				printf("%02hhx ", buff[i + j - base]);
		}
		printf(" ");
		for (j = 0; j < 16; j++) {
			if ((i + j < base) || (j >= len))
				printf(" ");
			else
				printf("%c", c_to_pc(buff[i + j - base]));
		}
		printf("\n");
	}
}

int cmd_inb(const char *args, int rel)
{
	unsigned long port;
	uint8_t value;
	int ret;

	ret = gen_parse_1ul(args, &port);
	if (ret)
		return parse_error(args);
		
	if (rel)
		port += iobase;
	ret = d_inb(port, &value);
	if (ret) {
		fprintf(stderr, "d_inb error: %s\n", strerror(-ret));
		return ret;
	}
	print_int8(value);
	return 0;
}

int cmd_outb(const char *args, int rel)
{
	unsigned long port, _value;
	uint8_t value;
	int ret;

	ret = gen_parse_2ul(args, &port, &_value);
	if (ret)
		return parse_error(args);
	if (rel)
		port += iobase;
	value = (uint8_t) _value;
	ret = d_outb(port, value);
	if (ret) {
		fprintf(stderr, "d_outb error: %s\n", strerror(-ret));
		return ret;
	}
	return 0;
}

int cmd_inw(const char *args, int rel)
{
	unsigned long port;
	uint16_t value;
	int ret;

	ret = gen_parse_1ul(args, &port);
	if (ret)
		return parse_error(args);
	if (rel)
		port += iobase;
	ret = d_inw(port, &value);
	if (ret) {
		fprintf(stderr, "d_inw error: %s\n", strerror(-ret));
		return ret;
	}
	print_int16(value);
	return 0;
}

int cmd_outw(const char *args, int rel)
{
	unsigned long port, _value;
	uint16_t value;
	int ret;

	ret = gen_parse_2ul(args, &port, &_value);
	if (ret)
		return parse_error(args);
	if (rel)
		port += iobase;
	value = (uint8_t) _value;
	ret = d_outb(port, value);
	if (ret) {
		fprintf(stderr, "d_outw error: %s\n", strerror(-ret));
		return ret;
	}
	return 0;
}

int cmd_inl(const char *args, int rel)
{
	unsigned long port;
	uint32_t value;
	int ret;

	ret = gen_parse_1ul(args, &port);
	if (ret)
		return parse_error(args);
	if (rel)
		port += iobase;
	ret = d_inl(port, &value);
	if (ret) {
		fprintf(stderr, "d_inl error: %s\n", strerror(-ret));
		return ret;
	}
	print_int32(value);
	return 0;
}

int cmd_outl(const char *args, int rel)
{
	unsigned long port, _value;
	uint32_t value;
	int ret;

	ret = gen_parse_2ul(args, &port, &_value);
	if (ret)
		return parse_error(args);
	if (rel)
		port += iobase;
	value = (uint8_t) _value;
	ret = d_outl(port, value);
	if (ret) {
		fprintf(stderr, "d_outl error: %s\n", strerror(-ret));
		return ret;
	}
	return 0;
}

int cmd_readb(const char *args, int rel)
{
	unsigned long addr;
	uint8_t value;
	int ret;

	ret = gen_parse_1ul(args, &addr);
	if (ret)
		return parse_error(args);
	if (rel)
		addr += membase;
	ret = d_readb(addr, &value);
	if (ret) {
		fprintf(stderr, "d_read error: %s\n", strerror(-ret));
		return ret;
	}
	print_int8(value);
	return 0;
}

int cmd_writeb(const char *args, int rel)
{
	unsigned long addr, _value;
	uint8_t value;
	int ret;

	ret = gen_parse_2ul(args, &addr, &_value);
	if (ret)
		return parse_error(args);
	if (rel)
		addr += membase;
	value = (uint8_t) _value;
	ret = d_writeb(addr, &value);
	if (ret) {
		fprintf(stderr, "d_write error: %s\n", strerror(-ret));
		return ret;
	}
	return 0;
}

int cmd_readw(const char *args, int rel)
{
	unsigned long addr;
	uint16_t value;
	int ret;

	ret = gen_parse_1ul(args, &addr);
	if (ret)
		return parse_error(args);
	if (rel)
		addr += membase;
	ret = d_readw(addr, &value);
	if (ret) {
		fprintf(stderr, "d_read error: %s\n", strerror(-ret));
		return ret;
	}
	print_int16(value);
	return 0;
}

int cmd_writew(const char *args, int rel)
{
	unsigned long addr, _value;
	uint16_t value;
	int ret;

	ret = gen_parse_2ul(args, &addr, &_value);
	if (ret)
		return parse_error(args);
	if (rel)
		addr += membase;
	value = (uint16_t) _value;
	ret = d_writew(addr, &value);
	if (ret) {
		fprintf(stderr, "d_write error: %s\n", strerror(-ret));
		return ret;
	}
	return 0;
}

int cmd_readl(const char *args, int rel)
{
	unsigned long addr;
	uint32_t value;
	int ret;

	ret = gen_parse_1ul(args, &addr);
	if (ret)
		return parse_error(args);
	if (rel)
		addr += membase;
	ret = d_readl(addr, &value);
	if (ret) {
		fprintf(stderr, "d_read error: %s\n", strerror(-ret));
		return ret;
	}
	print_int32(value);
	return 0;
}

int cmd_writel(const char *args, int rel)
{
	unsigned long addr, _value;
	uint32_t value;
	int ret;

	ret = gen_parse_2ul(args, &addr, &_value);
	if (ret)
		return parse_error(args);
	if (rel)
		addr += membase;
	value = (uint32_t) _value;
	ret = d_writel(addr, &value);
	if (ret) {
		fprintf(stderr, "d_write error: %s\n", strerror(-ret));
		return ret;
	}
	return 0;
}

int cmd_dump(const char *args, int align, int rel)
{
	unsigned long addr, len;
	uint8_t *buff;
	int ret;

	ret = gen_parse_2ul(args, &addr, &len);
	if (ret)
		return parse_error(args);
	if (rel)
		addr += membase;
	buff = (uint8_t *) malloc(len);
	if (buff == NULL) {
		fprintf(stderr, "debug: Memory allocation failure.\n");
		return -ENOMEM;
	}		
	ret = d_read(addr, buff, len);
	if (ret) {
		fprintf(stderr, "d_read error: %s\n", strerror(-ret));
		free(buff);
		return ret;
	}
	if (align)
		dump_area_a(addr, buff, len);
	else
		dump_area(addr, buff, len);
	free(buff);
	return 0;
}

int cmd_fill(const char *args, int rel)
{
	unsigned long addr, len, _value;
	uint8_t *buff;
	uint8_t value;
	int ret;

	ret = gen_parse_3ul(args, &addr, &len, &_value);
	if (ret)
		return parse_error(args);
	if (rel)
		addr += membase;
	value = (uint8_t) _value;
	buff = (uint8_t *) malloc(len);
	if (buff == NULL) {
		fprintf(stderr, "debug: Memory allocation failure.\n");
		return -ENOMEM;
	}
	memset(buff, value, len);	
	ret = d_write(addr, buff, len);
	free(buff);
	if (ret) {
		fprintf(stderr, "d_write error: %s\n", strerror(-ret));
		return ret;
	}
	return 0;
}

#ifdef HAVE_LIBPCI
static struct pci_access *pci_access;

int cmd_pci_inb(const char *args)
{
	struct pci_dev *dev;
	unsigned long port;
	u8 value;
	int ret;

	ret = gen_parse_1ul(args, &port);
	if (ret)
		return parse_error(args);

	if (pci_access == NULL) {
		fprintf(stderr, "PCI support is disabled\n");
		return -ENODEV;
	}

	dev = pci_get_dev(pci_access, pci_domain, pci_bus, pci_dev, pci_fn);
	if (dev == NULL) {
		fprintf(stderr, "pci_inb: %s.\n", strerror(ENODEV));
		return -ENODEV;
	}

	value = pci_read_byte(dev, port);
	print_int8(value);
	pci_free_dev(dev);
	return 0;
}

int cmd_pci_inw(const char *args)
{
	struct pci_dev *dev;
	unsigned long port;
	u16 value;
	int ret;

	ret = gen_parse_1ul(args, &port);
	if (ret)
		return parse_error(args);

	if (pci_access == NULL) {
		fprintf(stderr, "PCI support is disabled\n");
		return -ENODEV;
	}

	dev = pci_get_dev(pci_access, pci_domain, pci_bus, pci_dev, pci_fn);
	if (dev == NULL) {
		fprintf(stderr, "pci_inb: %s.\n", strerror(ENODEV));
		return -ENODEV;
	}

	value = pci_read_word(dev, port);
	print_int16(value);
	pci_free_dev(dev);
	return 0;
}

int cmd_pci_inl(const char *args)
{
	struct pci_dev *dev;
	unsigned long port;
	u32 value;
	int ret;

	ret = gen_parse_1ul(args, &port);
	if (ret)
		return parse_error(args);

	if (pci_access == NULL) {
		fprintf(stderr, "PCI support is disabled\n");
		return -ENODEV;
	}

	dev = pci_get_dev(pci_access, pci_domain, pci_bus, pci_dev, pci_fn);
	if (dev == NULL) {
		fprintf(stderr, "pci_inb: %s.\n", strerror(ENODEV));
		return -ENODEV;
	}

	value = pci_read_long(dev, port);
	print_int32(value);
	pci_free_dev(dev);
	return 0;
}

int cmd_pci_outb(const char *args)
{
	struct pci_dev *dev;
	unsigned long port, _value;
	u8 value;
	int ret;

	ret = gen_parse_2ul(args, &port, &_value);
	if (ret)
		return parse_error(args);
	value = (u8) _value;

	if (pci_access == NULL) {
		fprintf(stderr, "PCI support is disabled\n");
		return -ENODEV;
	}

	dev = pci_get_dev(pci_access, pci_domain, pci_bus, pci_dev, pci_fn);
	if (dev == NULL) {
		fprintf(stderr, "pci_inb: %s.\n", strerror(ENODEV));
		return -ENODEV;
	}

	pci_write_byte(dev, port, value);
	pci_free_dev(dev);
	return 0;
}

int cmd_pci_outw(const char *args)
{
	struct pci_dev *dev;
	unsigned long port, _value;
	u16 value;
	int ret;

	ret = gen_parse_2ul(args, &port, &_value);
	if (ret)
		return parse_error(args);
	value = (u16) _value;

	if (pci_access == NULL) {
		fprintf(stderr, "PCI support is disabled\n");
		return -ENODEV;
	}

	dev = pci_get_dev(pci_access, pci_domain, pci_bus, pci_dev, pci_fn);
	if (dev == NULL) {
		fprintf(stderr, "pci_inb: %s.\n", strerror(ENODEV));
		return -ENODEV;
	}

	pci_write_word(dev, port, value);
	pci_free_dev(dev);
	return 0;
}

int cmd_pci_outl(const char *args)
{
	struct pci_dev *dev;
	unsigned long port, _value;
	u32 value;
	int ret;

	ret = gen_parse_2ul(args, &port, &_value);
	if (ret)
		return parse_error(args);
	value = (u32) _value;

	if (pci_access == NULL) {
		fprintf(stderr, "PCI support is disabled\n");
		return -ENODEV;
	}

	dev = pci_get_dev(pci_access, pci_domain, pci_bus, pci_dev, pci_fn);
	if (dev == NULL) {
		fprintf(stderr, "pci_inb: %s.\n", strerror(ENODEV));
		return -ENODEV;
	}

	pci_write_long(dev, port, value);
	pci_free_dev(dev);
	return 0;
}
#else

int cmd_pci_unsupported(const char *args)
{
	fprintf(stderr, "PCI support is not compiled in. Make sure that "
			"libpci is installed.\n");
	return 0;
}

#define cmd_pci_inb(x)	cmd_pci_unsupported(x)
#define cmd_pci_inw(x)	cmd_pci_unsupported(x)
#define cmd_pci_inl(x)	cmd_pci_unsupported(x)
#define cmd_pci_outb(x)	cmd_pci_unsupported(x)
#define cmd_pci_outw(x)	cmd_pci_unsupported(x)
#define cmd_pci_outl(x)	cmd_pci_unsupported(x)
#endif

int cmd_rdmsr(const char *args)
{
	unsigned long port;
	uint64_t value;
	int ret;

	ret = gen_parse_1ul(args, &port);
	if (ret)
		return parse_error(args);
	ret = d_rdmsr(port, &value);
	if (ret) {
		fprintf(stderr, "d_rdmsr error: %s\n", strerror(-ret));
		return ret;
	}
	print_int64(value);
	return 0;
}

int cmd_wrmsr(const char *args)
{
	unsigned long port;
	unsigned long long value;
	uint64_t msr_value;
	int ret;

	ret = gen_parse_ul_ull(args, &port, &value);
	if (ret)
		return parse_error(args);
	msr_value = value;
	ret = d_wrmsr(port, &msr_value);
	if (ret) {
		fprintf(stderr, "d_wrmsr error: %s\n", strerror(-ret));
		return ret;
	}
	return 0;
}

int cmd_parambase(const char *args)
{
	unsigned long tmp;
	char cmd[64];
	int ret;

	ret = sscanf(args, "%63s %li", cmd, &tmp);
	if (ret < 2)
		return parse_error(args);
	if (tmp > 36) {
		fprintf(stderr, "Unsupported parambase: %ld.\n", tmp);
		return 1;
	}
	parambase = tmp;
	return 0;
}

int cmd_membase(const char *args)
{
	unsigned long addr;
	int ret;

	ret = gen_parse_1ul(args, &addr);
	if (ret)
		return parse_error(args);
	membase = addr;
	return 0;
}

int cmd_set_integer(const char *args, int *val)
{
	unsigned long addr;
	int ret;

	ret = gen_parse_1ul(args, &addr);
	if (ret)
		return parse_error(args);
	*val = addr;
	return 0;
}

int cmd_iobase(const char *args)
{
	unsigned long addr;
	int ret;

	ret = gen_parse_1ul(args, &addr);
	if (ret)
		return parse_error(args);
	iobase = addr;
	return 0;
}

static int parse_bool(const char *val)
{
	if (!strcmp(val, "yes") || !strcmp(val, "y") || !strcmp(val, "1"))
		return 1;
	return 0;
}

#define PARSE_BOOL_OPT(name, variable)					\
	do {								\
		if (!strcmp(opt, (name))) {				\
			(variable) = parse_bool(val);			\
			return 0;					\
		}							\
	} while (0)

int cmd_set(const char *args)
{
	char cmd[64], opt[64], val[64];
	int ret;

	ret = sscanf(args, "%63s %63s %63s", cmd, opt, val);
	if (ret < 3)
		return parse_error(args);
	PARSE_BOOL_OPT("arg_compat", ac_compat);
	PARSE_BOOL_OPT("dump_align", dump_align);
	PARSE_BOOL_OPT("rel_mem", rel_mem);
	PARSE_BOOL_OPT("rel_io", rel_io);
	PARSE_BOOL_OPT("hide_nonascii", hide_nonascii);
	PARSE_BOOL_OPT("hide_nonprintable", hide_nonprintable);
	return -EINVAL;
}

#undef PARSE_BOOL_OPT

int cmd_usleep(const char *args)
{
	char cmd[64];
	unsigned int arg;
	int ret;

	ret = sscanf(args, "%63s %d", cmd, &arg);
	if (ret < 2)
		return parse_error(args);
	usleep(arg);
	return 0;
}

static void show_info(void)
{
	fprintf(stderr, "debug "PACKAGE_VERSION"\n"
			"Copyright (C) 2005..2008  "
			"Krzysztof Mazur <krzysiek@podlesie.net>\n\n");
}
static void show_interactive_info(void)
{
	show_info();
	fprintf(stderr, "debug comes with ABSOLUTELY NO WARRANTY;\n"
			"This is free software, and you are welcome to "
			"redistribute it under\n"
			"certain conditions; type `license' for details.\n");
}

static void license(void)
{
	fprintf(stderr, "This program is free software; you can redistribute "
			"it and/or modify\n"
			"it under the terms of the GNU General Public License "
			"as published by\n"
			"the Free Software Foundation; either version 2 of the "
			"License, or\n"
			"(at your option) any later version.\n"
			"\n"
			"This program is distributed in the hope that it will "
			"be useful,\n"
			"but WITHOUT ANY WARRANTY; without even the implied "
			"warranty of\n"
			"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  "
			"See the\n"
			"GNU General Public License for more details.\n"
			"\n"
			"You should have received a copy of the GNU General "
			"Public License\n"
			"along with this program; if not, write to the Free "
			"Software\n"
			"Foundation, Inc., 59 Temple Place, Suite 330, Boston, "
			"MA  02111-1307  USA\n");
}

static void help(void)
{
	fprintf(stderr, "Supported commands:\n\n"
			"general: exit, q, quit, help, license, compat, native,"
			" parambase, set, usleep\n"
			"i/o: iobase\n"
			"i/o 8-bit: i, ib, inb, o, ob, outb,\n"
			"i/o 16-bit: iw, inw, ow, outw,\n"
			"i/o 32-bit: il, inl, ol, outl,\n"
			"memory: membase\n"
			"memory 8-bit: r, rb, readb, w, wb, writeb,\n"
			"memory 16-bit: rw, readw, ww, writew,\n"
			"memory 32-bit: rl, readl, wl, writel,\n"
			"memory block: d, dump, dn, dumpn, da, dumpa, f, fill"
			"\n");
}

static void show_usage(char *name)
{
	show_info();
	fprintf(stderr,	"usage: %s [options] (file)\n"
			"Options:\n"
			"  -c, --compat		Linux MiniDEBUG compatibility "
			"mode\n"
			"  -i, --io		set io access method\n"
			"  -h, --help		show this help\n"
			"      --noreadline	don't use readline\n"
			"      --readline	use readline\n"
			"  -q, --quiet		quiet mode\n"
			"  -V, --version		show version\n",
			(name == NULL) ? "debug" : name);
	exit(0);
}

static void enable_compat(void)
{
	ac_compat = 1;
	dump_align = 1;
	rel_mem = 1;
	rel_io = 1;
	hide_nonascii = 0;
	hide_nonprintable = 1;
}

static void enable_native(void)
{
	ac_compat = 0;
	dump_align = 0;
	rel_mem = 0;
	rel_io = 0;
	hide_nonascii = 1;
	hide_nonprintable = 1;
}

static void parse_command_line(int argc, char **args)
{
	int c;
	int option_index = 0;

	static struct option long_options[] = {
		{"compat", 0, 0, 'c'},
		{"help", 0, 0, 'h'},
		{"io", 1, 0, 'i'},
		{"noreadline", 0, 0, 2},
		{"quiet", 0, 0, 'q'},
		{"readline", 0, 0, 1},
		{"version", 0, 0, 'V'},
		{0, 0, 0, 0},
	};
	
	while (1) {
		c = getopt_long(argc, args, "ci:hqV", long_options,
				&option_index);
		if (c == -1)
			break;
		switch (c) {
			case 1:
				use_readline = USE_READLINE_YES;
				break;
			case 2:
				use_readline = USE_READLINE_NO;
				break;
			case 'c':
				enable_compat();
				break;
			case 'i':
				if (!strcmp(optarg, "ioperm"))
					io_access_mode = IOA_IOPERM;
				else if (!strcmp(optarg, "devport"))
					io_access_mode = IOA_DEV_PORT;
				else if (!strcmp(optarg, "dummy"))
					io_access_mode = IOA_DUMMY;
				else {
					fprintf(stderr, "Unsupported i/o port "
							"access mode: %s\n",
							optarg);
				}
				break;
			case 'h':
				show_usage(args[0]);
				break;
			case 'q':
				quiet_mode = 1;
				break;
			case 'V':
				exit(0);
				break;
			default:
				fprintf(stderr, "Unknown option.\n");
				exit(1);
		}
	}

	if (optind < argc) {
		run_script = args[optind++];
		if (optind < argc)
			fprintf(stderr, "warning: unknown parameters detected."
					"\n");	
	}
}

#ifdef HAVE_LIBREADLINE
static char *history_file = NULL;

static void init_history(void)
{
	char *home = getenv("HOME");

	using_history();
	stifle_history(1000);
	history_file = malloc(strlen(home) + 100);
	if (history_file != NULL) {
		sprintf(history_file, "%s/.debug_history", home);
		read_history(history_file);
	}
}

#else

static void init_history(void)
{
}

#endif

static void prepare_exit(void)
{
#ifdef HAVE_LIBREADLINE
	if (history_file != NULL)
		write_history(history_file);
#endif
}

int main(int argc, char **args)
{
	char *tmp = NULL;
	char cmd[64];
	int ret;

	parse_command_line(argc, args);
	if (!quiet_mode && (run_script == NULL))
		show_interactive_info();
	d_ioinit(io_access_mode);
	d_meminit();
	d_msrinit(0);
#ifdef HAVE_LIBPCI
	pci_access = pci_alloc();
	if (pci_access == NULL)
		fprintf(stderr, "pci_alloc() failed.\n");
	pci_init(pci_access);
#endif
#ifdef HAVE_LIBREADLINE
	rl_outstream = stderr;
#endif
	if (run_script != NULL) {
		if (freopen(run_script, "r", stdin) == NULL) {
			fprintf(stderr, "Can't open %s.\n", run_script);
			exit(1);
		}
		if (use_readline == USE_READLINE_DEFAULT)
			use_readline = 0;
		quiet_mode = 1;
	} else {
		init_history();
		atexit(prepare_exit);
	}
	while (1) {
		if (tmp)
			free(tmp);
		if (use_readline)
			tmp = rl_gets();
		else
			tmp = std_gets();
		if (tmp == NULL) {
			fprintf(stderr, "\n");
			break;
		}
		if ((*tmp == '\0') || (*tmp == '#'))
			continue;
		ret = sscanf(tmp, "%63s\n", cmd);
		if ((ret < 1) || (*cmd == '\0'))
			continue;
		if (!strcmp(cmd, "exit") || !strcmp(cmd, "quit") ||
				!strcmp(cmd, "q"))
			break;
		if (!strcmp(cmd, "i") || !strcmp(cmd, "ib") ||
				!strcmp(cmd, "inb")) {
			cmd_inb(tmp, rel_io);
			continue;
		}
		if (!strcmp(cmd, "o") || !strcmp(cmd, "ob") ||
				!strcmp(cmd, "outb")) {
			cmd_outb(tmp, rel_io);
			continue;
		}
		if (!strcmp(cmd, "iw") || !strcmp(cmd, "inw")) {
			cmd_inw(tmp, rel_io);
			continue;
		}
		if (!strcmp(cmd, "ow") || !strcmp(cmd, "outw")) {
			cmd_outw(tmp, rel_io);
			continue;
		}
		if (!strcmp(cmd, "il") || !strcmp(cmd, "inl")) {
			cmd_inl(tmp, rel_io);
			continue;
		}
		if (!strcmp(cmd, "ol") || !strcmp(cmd, "outl")) {
			cmd_outl(tmp, rel_io);
			continue;
		}
		if (!strcmp(cmd, "r") || !strcmp(cmd, "rb") ||
				!strcmp(cmd, "readb")) {
			cmd_readb(tmp, rel_mem);
			continue;
		}
		if (!strcmp(cmd, "w") || !strcmp(cmd, "wb") ||
				!strcmp(cmd, "writeb")) {
			cmd_writeb(tmp, rel_mem);
			continue;
		}
		if (!strcmp(cmd, "ww") || !strcmp(cmd, "writew")) {
			cmd_writew(tmp, rel_mem);
			continue;
		}
		if (!strcmp(cmd, "rw") || !strcmp(cmd, "readw")) {
			cmd_readw(tmp, rel_mem);
			continue;
		}
		if (!strcmp(cmd, "wl") || !strcmp(cmd, "writel")) {
			cmd_writel(tmp, rel_mem);
			continue;
		}
		if (!strcmp(cmd, "rl") || !strcmp(cmd, "readl")) {
			cmd_readl(tmp, rel_mem);
			continue;
		}
		if (!strcmp(cmd, "d") || !strcmp(cmd, "dump")) {
			cmd_dump(tmp, dump_align, rel_mem);
			continue;
		}
		if (!strcmp(cmd, "da") || !strcmp(cmd, "dumpa")) {
			cmd_dump(tmp, 1, rel_mem);
			continue;
		}
		if (!strcmp(cmd, "dn") || !strcmp(cmd, "dumpn")) {
			cmd_dump(tmp, 0, rel_mem);
			continue;
		}
		if (!strcmp(cmd, "f") || !strcmp(cmd, "fill")) {
			cmd_fill(tmp, rel_mem);
			continue;
		}
		if (!strcmp(cmd, "set")) {
			cmd_set(tmp);
			continue;
		}
		if (!strcmp(cmd, "membase")) {
			cmd_membase(tmp);
			continue;
		}
		if (!strcmp(cmd, "iobase")) {
			cmd_iobase(tmp);
			continue;
		}
		if (!strcmp(cmd, "license")) {
			license();
			continue;
		}
		if (!strcmp(cmd, "help")) {
			help();
			continue;
		}
		if (!strcmp(cmd, "compat")) {
			enable_compat();
			continue;
		}
		if (!strcmp(cmd, "native")) {
			enable_native();
			continue;
		}
		if (!strcmp(cmd, "parambase")) {
			cmd_parambase(tmp);
			continue;
		}
		if (!strcmp(cmd, "usleep")) {
			cmd_usleep(tmp);
			continue;
		}
		if (!strcmp(cmd, "pci_domain")) {
			cmd_set_integer(tmp, &pci_domain);
			continue;
		}
		if (!strcmp(cmd, "pci_bus") || !strcmp(cmd, "pbus")) {
			cmd_set_integer(tmp, &pci_bus);
			continue;
		}
		if (!strcmp(cmd, "pci_dev") || !strcmp(cmd, "pdev")) {
			cmd_set_integer(tmp, &pci_dev);
			continue;
		}
		if (!strcmp(cmd, "pci_fn") || !strcmp(cmd, "pfn")) {
			cmd_set_integer(tmp, &pci_fn);
			continue;
		}
		if (!strcmp(cmd, "pci_inb") || !strcmp(cmd, "pinb")
				|| !strcmp(cmd, "pi")) {
			cmd_pci_inb(tmp);
			continue;
		}
		if (!strcmp(cmd, "pci_inw") || !strcmp(cmd, "pinw")) {
			cmd_pci_inw(tmp);
			continue;
		}
		if (!strcmp(cmd, "pci_inl") || !strcmp(cmd, "pinl")) {
			cmd_pci_inl(tmp);
			continue;
		}
		if (!strcmp(cmd, "pci_outb") || !strcmp(cmd, "poutb")
				|| !strcmp(cmd, "po")) {
			cmd_pci_outb(tmp);
			continue;
		}
		if (!strcmp(cmd, "pci_outw") || !strcmp(cmd, "poutw")) {
			cmd_pci_outw(tmp);
			continue;
		}
		if (!strcmp(cmd, "pci_outl") || !strcmp(cmd, "poutl")) {
			cmd_pci_outl(tmp);
			continue;
		}
		if (!strcmp(cmd, "rdmsr")) {
			cmd_rdmsr(tmp);
			continue;
		}
		if (!strcmp(cmd, "wrmsr")) {
			cmd_wrmsr(tmp);
			continue;
		}
		if (!strcmp(cmd, "!")) {
			system(tmp + 1);
			continue;
		}
		fprintf(stderr, "Unknown command: %s\n", cmd);
	}
	if (!quiet_mode)
		fprintf(stderr, "Bye\n");
	return 0;
}
